# have to monkey patch to work with WSL as workaround for https://bugs.python.org/issue38633
import errno, shutil
orig_copyxattr = shutil._copyxattr
def patched_copyxattr(src, dst, *, follow_symlinks=True):
	try:
		orig_copyxattr(src, dst, follow_symlinks=follow_symlinks)
	except OSError as ex:
		if ex.errno != errno.EACCES: raise
shutil._copyxattr = patched_copyxattr


from ignite.engine.events import Events, CallableEventWithFilter
import re
def event_argument(event):
    if not event:
        return None
    elif isinstance(event, (Events, CallableEventWithFilter)):
        return event
    elif isinstance(event, str):
        m = re.match("^([A-Z_]+)(\\(.*\\))?$", event)
        if not m:
            raise ValueError("Invalid event format: {}. Should be something like \"ITERATION_COMPLETED(every=100)\".".format(event))
        evt = m.group(1)
        if not hasattr(Events, evt):
            raise ValueError("Invalid event name: {}".format(evt))
        evt = getattr(Events, evt)
        filt = m.group(2)
        if not filt:
            return evt
        else:
            m = re.match("\\((.*)=(.*)\\)", filt)
            k = m.group(1)
            v = int(m.group(2))
            kwargs = {k:v}
            return evt(**kwargs)
    else:
        raise ValueError("Event should be string or ignite.engine.events.Event but is {}".format(type(event)))

if __name__=='__main__':
    print(event_argument(None))
    print(event_argument(""))
    print(event_argument(Events.ITERATION_COMPLETED))
    print(event_argument(Events.ITERATION_COMPLETED(every=10)))
    print(event_argument("ITERATION_COMPLETED"))
    print(event_argument("ITERATION_COMPLETED(every=10)"))




class IgniterConfig(object):
    def __init__(
            self,
            model_args=None,
            train_args=None,
            eval_args=None,
            description=None,
            train_inputs=None,
            eval_inputs=None,
            make_model=None,
            make_evaluator=None,
            make_trainer=None,
            inference_spec=None,
            **igniter_args
    ):
        self.model_args = model_args
        self.train_args = train_args
        self.eval_args = eval_args
        self.description = description or "Experiment CLI"
        self.train_inputs = train_inputs or {}
        self.eval_inputs = eval_inputs or {}
        self.make_model = make_model
        self.make_evaluator = make_evaluator
        self.make_trainer = make_trainer
        self.inference_spec = inference_spec
        self.igniter_args = igniter_args



import torch
from torch import nn


class MnistModel(nn.Module):
    def __init__(self):
        super(MnistModel, self).__init__()
        self.model = nn.Sequential(
            nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, stride=2),
            nn.LeakyReLU(),
            nn.Conv2d(in_channels=32, out_channels=64,
                      kernel_size=3, stride=2),
            nn.LeakyReLU(),
            nn.Conv2d(in_channels=64, out_channels=128,
                      kernel_size=3, stride=1),
            nn.LeakyReLU(),
            nn.Conv2d(in_channels=128, out_channels=10,
                      kernel_size=3, stride=1),
        )

    def forward(self, input):
        return torch.mean(self.model(input), dim=(2, 3))




import torch
from .image_input_fn import input_fn as image_input_fn


def input_fn(request_body, request_content_type):
    """
    Convert input for your model
    """
    print("Running greyscale_image_input_fn")
    tensor = image_input_fn(request_body, request_content_type)
    if tensor.size(0) > 1:
        tensor = torch.mean(tensor, dim=0, keepdim=True)
    return tensor



from torch.utils.data import IterableDataset, get_worker_info, DataLoader
from aws_sagemaker_remote.util.pipes import chunk_iterable, ProtobufPipeIterator, epoch_iterable
from sagemaker.amazon.common import read_recordio
import os
from torch.multiprocessing import Value, Lock, Queue, spawn, Process, get_context, Event, JoinableQueue
from queue import Empty, Full

MAX_FAIL = 30


def get_worker():
    info = get_worker_info()
    if not info:
        worker = 0
    else:
        worker = info.id
    assert worker >= 0
    return worker


class MultiprocessingSource(object):
    def __init__(self, iterable, maxsize, epochs=0, multiprocessing_context=None):
        self.iterable = iterable
        self.epochs = epochs
        self.maxsize = maxsize
        self.multiprocessing_context = multiprocessing_context or get_context()

    def run(self):
        #print("Running process")
        epochs = epoch_iterable(self.epochs)
        done = False
        sent = 0
        for _ in epochs:
            if done:
                break
            for example in self.iterable:
                if done:
                    break
                # print("Queueing item: {}".format({k: v.size()
                #                                  for k, v in example.items()}))
                failures = 0
                for v in example.values():
                    assert not v.requires_grad
                # if sent > self.maxsize:

                # todo: check performance gains
                """
                try:
                    buf = self.return_queue.get(block=False)
                    for k, v in example.items():
                        buf[k].copy_(v)
                    example = buf
                except Empty as e:
                    pass
                    #print("Return queue is empty")
                    #raise Exception("Return queue is empty") from e
                """
                while True:
                    if self.request_close.value:
                        #print("Got flag")
                        done = True
                        break
                    try:
                        self.queue.put(example, timeout=1)
                        sent += 1
                        break
                    except Full as e:
                        failures += 1
                        if failures >= MAX_FAIL:
                            raise Exception("Putting on queue failed") from e
                        else:
                            continue
        self.is_done.value = True
        self.done_event.wait(timeout=120)
        #print("Exiting process")

    def open(self):
        #print("Spawning process")
        self.queue = self.multiprocessing_context.Queue(
            maxsize=self.maxsize)
        # self.return_queue = self.multiprocessing_context.Queue(
        #    maxsize=self.maxsize)
        self.is_done = self.multiprocessing_context.Value('b')
        self.is_done.value = False
        self.request_close = self.multiprocessing_context.Value('b')
        self.request_close.value = False
        self.done_event = self.multiprocessing_context.Event()
        ctx = self.multiprocessing_context.Process(
            target=MultiprocessingSource.run,
            args=(self,),
            # nprocs=1,
            # join=False,  # True,
            # daemon=True,
            # start_method='spawn'
        )
        # print(ctx)
        self.ctx = ctx
        self.ctx.start()
        return ctx

    def close(self):
        self.request_close.value = True
        #self.is_done.value = True
        self.done_event.set()
        # self.queue.join()
        self.ctx.join()

    def __iter__(self):
        failures = 0
        while True:
            rec = None
            # if self.request_close.value:
            #    break
            try:
                rec = self.queue.get(timeout=1)
            except Empty:
                rec = None
            if rec is None:
                if self.is_done.value or failures >= MAX_FAIL:
                    break
                else:
                    failures += 1
                    continue
            #rec = self.transform(rec, worker)
            for v in rec.values():
                assert not v.requires_grad
            yield rec
            # self.queue.task_done()
            #self.return_queue.put(rec, block=False)
        if failures >= MAX_FAIL:
            raise Exception(
                f"Ran {MAX_FAIL} timeout failures without seeing is_done flag")


class PipeDataset(IterableDataset):
    def get_worker(self):
        return get_worker()

    # def transform(self, rec):
    #    return rec

    def __init__(
        self,
        path,
        num_workers,
        pipe_cls=ProtobufPipeIterator,
        multiprocessing_context=None,
        **pipe_args
    ):
        if isinstance(path, str):
            path = [path]
        self.path = path
        self.pipes = [
            pipe_cls(
                path=p,
                **pipe_args
            )
            for p in self.path
        ]
        self.mw = None
        if num_workers > 1:
            if len(path) == 1:
                self.mw = MultiprocessingSource(
                    self.pipes[0],
                    maxsize=20,  # todo,
                    epochs=1,
                    multiprocessing_context=multiprocessing_context
                )
            else:
                assert len(path) == num_workers
        else:
            assert len(path) == 1

    def __iter__(self):
        worker = self.get_worker()
        if self.mw:
            it = self.mw
        else:
            it = self.pipes[worker]
        return iter(it)
        # for i in it:
        #    yield i #self.transform(i)

    def open(self):
        if self.mw:
            self.mw.open()

    def close(self):
        if self.mw:
            self.mw.close()


class MultiprocessingCounter(object):
    def __init__(self):
        self.value = Value('i', 0)
        self.lock = Lock()
        self.local = 0


class ClosingDataLoader(DataLoader):
    def __iter__(self):
        self.dataset.open()
        for rec in super(ClosingDataLoader, self).__iter__():
            yield rec
        self.dataset.close()




from collections import OrderedDict
import torch
from ignite.engine import Events
import os
import torchvision.utils as vutils
import re
from tqdm import tqdm
from ignite.handlers import ModelCheckpoint, Timer
from ignite.metrics import RunningAverage
from torch import nn
import signal
from contextlib import contextmanager
import numpy as np
import warnings

EVAL_MESSAGE = "[{epoch}/{max_epochs}][{i}/{max_i}][Evaluation]"
TRAIN_MESSAGE = "[{epoch}/{max_epochs}][{i}/{max_i}]"
INTERRUPTED = "KeyboardInterrupt caught. Exiting gracefully."


class StateDictStorage(nn.Module):
    def __init__(self):
        super(StateDictStorage, self).__init__()
        self._state_dict = None

    def load_state_dict(self, state_dict, strict=True):
        self._state_dict = state_dict

    def state_dict(self):
        return self._state_dict


def get_metrics(engine, metric_names='all'):
    if metric_names == 'all':
        metrics = engine.state.metrics
    else:
        metrics = OrderedDict([
            (metric_name, engine.state.metrics[metric_name])
            for metric_name in metric_names
        ])
    metrics = tensors_to_numpy(metrics)
    metrics = OrderedDict([
        (k, np.array(v).item())
        for k, v
        in metrics.items()
        if np.array(v).size == 1
    ])
    return metrics


def print_logs(engine, trainer=None, fmt=TRAIN_MESSAGE, metric_fmt=" | {name}: {value}",
               metric_names='all'):
    if trainer is None:
        trainer = engine
    message = fmt.format(
        epoch=trainer.state.epoch,
        max_epochs=trainer.state.max_epochs,
        i=engine.state.iteration -
        ((engine.state.epoch-1)*engine.state.epoch_length),
        max_i=engine.state.epoch_length
    )
    for name, value in get_metrics(engine, metric_names=metric_names).items():
        message += metric_fmt.format(
            name=name, value=str(round(value, 3)))
    tqdm.write(message)


def save_logs(engine, fname, trainer=None, metric_names='all'):
    if trainer is None:
        trainer = engine
    columns = ['iteration', 'epoch']
    values = [str(trainer.state.iteration), str(trainer.state.epoch)]
    for key, value in get_metrics(engine, metric_names=metric_names).items():
        columns.append(key)
        values.append(str(round(value, 5)))
    with open(fname, "a") as f:
        if f.tell() == 0:
            print("\t".join(columns), file=f)
        print("\t".join(values), file=f)


def create_plots(engine, logs_fname, plots_fname, metric_names='all'):
    try:
        import matplotlib as mpl

        mpl.use("agg")

        import numpy as np
        import pandas as pd
        import matplotlib.pyplot as plt

    except ImportError:
        warnings.warn(
            "Loss plots will not be generated -- pandas or matplotlib not found")

    else:
        if os.path.exists(logs_fname):
            df = pd.read_csv(logs_fname,
                             delimiter="\t", index_col="iteration")
            if metric_names != 'all':
                df = df.filter(
                    items=metric_names,
                    axis='columns'
                )
            _ = df.plot(subplots=True, figsize=(20, 20))
            _ = plt.xlabel("Iteration number")
            fig = plt.gcf()
            fig.savefig(plots_fname)
            plt.close(fig)


# todo: add sigint handling
def handle_exception(engine, e, callback=None, **kwargs):
    if isinstance(e, KeyboardInterrupt):
        engine.terminate()
        tqdm.write(INTERRUPTED)
        if callback is not None:
            callback(engine, **kwargs)
    else:
        raise e


def chain_callbacks(callbacks=None, **kwargs):
    if callbacks is not None:
        if isinstance(callbacks, (list, tuple)):
            for callback in callbacks:
                if callback is not None:
                    callback(**kwargs)
        else:
            callbacks(**kwargs)


def kill_signals():
    signals = [signal.SIGINT, signal.SIGTERM]
    if hasattr(signal, 'SIGHUP'):
        signals.append(signal.SIGHUP)
    return signals


@contextmanager
def capture_signals(signals=None, callback=None, die=False, **kwargs):
    if signals is None:
        signals = kill_signals()
    original_handlers = [signal.getsignal(sig) for sig in signals]

    def handle_signal(sig, frame):
        if callback is not None:
            callback(**kwargs)
        raise KeyboardInterrupt("Received signal {}".format(sig))

    for sig in signals:
        signal.signal(sig, handle_signal)
    try:
        yield
    except KeyboardInterrupt as e:
        tqdm.write(str(e))
        if die:
            raise e
    finally:
        # note: only works if old handler was originally installed by Python
        for sig, original_handler in zip(signals, original_handlers):
            signal.signal(sig, original_handler)


def load_from(model_dir, to_load):
    path, iteration = find_last_checkpoint(model_dir)
    if path is None:
        raise FileNotFoundError("No checkpoints found in {}".format(model_dir))
    loaded = torch.load(path)
    for k, v in to_load.items():
        v.load_state_dict(loaded[k])
    return path, iteration


def find_last_checkpoint(output_dir):
    checkpoint_handler = ModelCheckpoint(
        output_dir, filename_prefix="", require_empty=False)
    return get_last_checkpoint(checkpoint_handler=checkpoint_handler)


def get_last_checkpoint(checkpoint_handler: ModelCheckpoint):
    dirname = checkpoint_handler.save_handler.dirname
    if hasattr(checkpoint_handler, '_fname_prefix'):
        filename_prefix = checkpoint_handler._fname_prefix
    elif hasattr(checkpoint_handler, "filename_prefix"):
        filename_prefix = checkpoint_handler.filename_prefix
    else:
        filename_prefix = ""
    if hasattr(checkpoint_handler, '_ext'):
        ext = checkpoint_handler._ext
    elif hasattr(checkpoint_handler, "ext"):
        ext = checkpoint_handler.ext
    else:
        ext = "pt"
    if not ext.startswith("."):
        ext = "\\.{}".format(ext)
    fmt = "{}checkpoint_(\\d+){}".format(filename_prefix, ext)

    def parse(fn):
        m = re.match(fmt, fn)
        if m:
            return int(m.group(1))
        else:
            return None
    files = list((fn, parse(fn)) for fn in os.listdir(dirname))
    files = filter(lambda x: x[1] is not None, files)
    files = sorted(files, key=lambda x: x[1])
    if files:
        last_file, last_iteration = files[-1]
        return os.path.join(dirname, last_file), last_iteration
    else:
        return None, None


def get_value(key):
    def fn(x):
        return x[key]
    return fn


def get_mean_value(key):
    def fn(x):
        return torch.mean(x[key])
    return fn


def auto_metric(value, cls=RunningAverage):
    if isinstance(value, str):
        return cls(output_transform=get_value(value))
    else:
        return value


def image_saver(engine, output_path, fn):
    path = output_path.format(engine.state.epoch)
    vutils.save_image(fn(engine).detach(), path)


def image_saver_callback(output_path, images):
    """
    Add a callback to save images

    """
    def fn(engine):
        for key, value in images.items():
            engine.add_event_handler(
                event_name=Events.EPOCH_COMPLETED,
                handler=image_saver,
                output_path=os.path.join(output_path, key),
                fn=value
            )
    return fn


def apply_to_tensors(tensors, fn):
    if torch.is_tensor(tensors):
        return fn(tensors)
    elif isinstance(tensors, list):
        return list(apply_to_tensors(tensor, fn) for tensor in tensors)
    elif isinstance(tensors, tuple):
        return tuple(apply_to_tensors(tensor, fn) for tensor in tensors)
    elif isinstance(tensors, dict):
        return OrderedDict((k, apply_to_tensors(tensor, fn)) for k, tensor in tensors.items())
    else:
        return tensors


def tensors_to_device(device):
    def fn(tensors):
        return apply_to_tensors(tensors=tensors, fn=lambda tensor: tensor.to(device))
    return fn


def tensors_to_items(tensors):
    return apply_to_tensors(tensors=tensors, fn=lambda tensor: tensor.item())


def tensors_to_numpy(tensors):
    return apply_to_tensors(tensors=tensors, fn=lambda tensor: tensor.detach().cpu().numpy())


def timer_metric(engine, name='timer'):
    timer = Timer(average=True)
    timer.attach(
        engine,
        start=Events.EPOCH_STARTED,
        resume=Events.ITERATION_STARTED,
        pause=Events.ITERATION_COMPLETED,
        step=Events.ITERATION_COMPLETED)

    def handler(_engine):
        _engine.state.metrics[name] = timer.value()
    engine.add_event_handler(
        event_name=Events.ITERATION_COMPLETED,
        handler=handler
    )


if __name__ == '__main__':
    with capture_signals():
        import time
        time.sleep(200)
    print("Done")



import json
import torch
import numpy as np


class TensorEncoder(json.JSONEncoder):
    def default(self, obj):
        if torch.is_tensor(obj):
            obj = obj.detach().cpu().numpy()
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def output_fn(prediction, content_type):
    """
    Convert model output
    """
    print("Running json_output_fn")
    if content_type == 'application/json':
        return json.dumps(prediction, cls=TensorEncoder), content_type
    else:
        raise ValueError("Unknown content type: {}".format(content_type))




import numpy
from io import BytesIO
import torch
from scipy.io import wavfile
import json
from urllib.parse import urlparse
import boto3
import os

def input_fn(request_body, request_content_type):
    request_content_type = (request_content_type or "").strip()
    if request_content_type in ['audio/wav', 'audio/wave', 'audio/x-wav']:
        fs, data = wavfile.read(BytesIO(request_body))
        #data = torch.from_numpy(data)
        return fs, data
    elif request_content_type in ['application/json','text/javascript','application/javascript','text/json']:
        if isinstance(request_body, bytes):
            request_body = request_body.decode('utf-8')
        info = json.loads(request_body)
        if 's3' in info:
            uri = info['s3']
            url = urlparse(uri)
            bucket = url.netloc
            ext = os.path.splitext(uri)[-1]
            ext = ext.lstrip(".")
            key = url.path.lstrip("/")
            session = boto3.Session()
            s3 = session.client('s3')
            obj = s3.get_object(Bucket=bucket, Key=key)
            contents = BytesIO(obj['Body'].read())
            if ext.lower() == 'wav':
                fs, data = wavfile.read(contents)
                return fs, data
            else:
                raise ValueError("s3 object [{}] unknown extension [{}]".format(uri, ext))
        elif 'local' in info:
            path = info['local']
            if not os.path.exists(path):
                raise FileNotFoundError("File [{}] not found".format(path))
            fs, data = wavfile.read(path)
            return fs, data
        else:
            raise ValueError('JSON requests should include an `s3` or `local` key')  
    else:
        raise ValueError(
            "Unsupported content type: \"{}\"".format(request_content_type))



import argparse
import contextlib
import os
import random
import warnings
import numpy as np
from ignite.contrib.handlers.mlflow_logger import MLflowLogger
import mlflow
from .mlflow_ctx import mlflow_ctx, get_mlflow_logger
import re
from aws_sagemaker_remote.modules import module_path
import shutil
from .events import event_argument
import torch
import torch.nn as nn
import torch.optim as optim
from tqdm import tqdm
import torch.utils.data as data
import pytorch_igniter.inference.inference
import json
from ignite.engine import Engine, Events
from ignite.handlers import ModelCheckpoint, Timer
from ignite.metrics import RunningAverage
from pytorch_igniter.metrics import SafeAverage

import torchvision.datasets as dset
import torchvision.transforms as transforms
from torch.autograd import backward
import yaml
from .spec import RunSpec
from .engine import build_engine
from .util import handle_exception, get_last_checkpoint, get_metrics, capture_signals
from .ssm import get_secret
import re
import pytorch_igniter.copyxattr_patch


def python_ignore(path, names):
    return [
        name for name in names
        if re.match("__pycache__|\\.git", name)
    ]


def export_code(model_dir, inference_args, inference_spec):
    os.makedirs(model_dir, exist_ok=True)
    if inference_args:
        with open(os.path.join(model_dir, 'args.json'), 'w') as f:
            json.dump(vars(inference_args), f)
    if inference_spec:
        # with open(os.path.join(model_dir,'config.json'),'w') as f:
        #    json.dump({
        #        "inferencer_module": inference_spec.inferencer.__class__.__module__,
        #        "inferencer_class": inference_spec.r.__class__.__name__,
        #    }, f)
        code_dir = os.path.join(model_dir, 'code')
        os.makedirs(code_dir, exist_ok=True)
        for dep in inference_spec.dependencies:
            dep = module_path(dep)
            des = os.path.join(
                code_dir, os.path.basename(dep)
            )
            if os.path.exists(des):
                shutil.rmtree(des)
            shutil.copytree(
                dep,
                des,
                # dirs_exist_ok=True,
                ignore=python_ignore
            )
        if inference_spec.requirements:
            shutil.copyfile(inference_spec.requirements, os.path.join(
                code_dir, 'requirements.txt'
            ))
        inference_py = os.path.join(code_dir, 'inference.py')
        shutil.copyfile(
            pytorch_igniter.inference.inference.__file__,
            inference_py
        )
        with open(inference_py, 'a') as f:
            f.write("\n")
            f.write("\n")
            from types import FunctionType
            if isinstance(inference_spec.input_fn, str):
                f.write("{}\n".format(inference_spec.input_fn))
            else:
                f.write("from {} import {} as input_fn\n".format(
                    inference_spec.input_fn.__module__,
                    inference_spec.input_fn.__name__
                ))
            if isinstance(inference_spec.output_fn, str):
                f.write("{}\n".format(inference_spec.output_fn))
            else:
                f.write("from {} import {} as output_fn\n".format(
                    inference_spec.output_fn.__module__,
                    inference_spec.output_fn.__name__
                ))
            if isinstance(inference_spec.inferencer, str):
                f.write("{}\n".format(inference_spec.inferencer))
            else:
                f.write("from {} import {} as inferencer_fn\n".format(
                    inference_spec.inferencer.__module__,
                    inference_spec.inferencer.__name__
                ))


def export_model(model_dir, model):
    os.makedirs(model_dir, exist_ok=True)
    torch.save(
        # {k: v.state_dict() for k, v in to_save.items()},
        {"model": model.state_dict()},
        os.path.join(model_dir, 'model.pt')
    )


def export_all(model_dir, model, inference_args, inference_spec):
    export_code(
        model_dir=model_dir,
        inference_args=inference_args,
        inference_spec=inference_spec)
    export_model(
        model_dir=model_dir,
        model=model)



import os
from argparse import Namespace
import importlib
import json
import torch

"""
Runs inference on SageMaker
"""


def model_fn(model_dir):
    """
    Load your model from model_dir
    """
    print("Running model_fn: {}".format(model_dir))
    # with open(os.path.join(model_dir, 'config.json'), 'r') as f:
    #    config = json.load(f)
    with open(os.path.join(model_dir, 'args.json'), 'r') as f:
        vargs = json.load(f)
    args = Namespace(**vargs)
    # inferencer_module = importlib.import_module(
    #    config['inferencer_module']
    # )
    #inferencer_class = getattr(inferencer_module, config['inferencer_class'])
    inferencer = inferencer_fn(args)
    model = inferencer.model
    loaded = torch.load(os.path.join(model_dir, 'model.pt'))
    model.load_state_dict(loaded['model'])
    return inferencer


def predict_fn(input_data, model):
    """
    Run your model
    """
    print("Running predict_fn")
    return model.inference(input_data)
